import { Slicer, DayGridRange, DayTableModel, DateRange, DateProfileGenerator, DateEnv, BaseComponent, ViewProps, DateMarker, DateProfile, DateFormatter, ViewContext, DayTableCell, Hit, EventRangeProps, EventSegUiInteractionState, DateComponent, ViewOptionsRefined, RefMap, SlicedCoordRange } from '@fullcalendar/core/internal';
import { createElement, Ref, VNode } from '@fullcalendar/core/preact';
import { CustomContentGenerator, ClassNamesGenerator, DidMountHandler, WillUnmountHandler, DayHeaderContentArg, CssDimValue } from '@fullcalendar/core';

declare class DayTableSlicer extends Slicer<DayGridRange, [DayTableModel]> {
    forceDayIfListItem: boolean;
    sliceRange(dateRange: DateRange, dayTableModel: DayTableModel): DayGridRange[];
}

declare class TableDateProfileGenerator extends DateProfileGenerator {
    buildRenderRange(currentRange: any, currentRangeUnit: any, isRangeAllDay: any): DateRange;
}
declare function buildDayTableRenderRange(props: {
    currentRange: DateRange;
    snapToWeek: boolean;
    fixedWeekCount: boolean;
    dateEnv: DateEnv;
}): DateRange;

declare class DayGridView extends BaseComponent<ViewProps> {
    private buildDayTableModel;
    private buildDateRowConfigs;
    private createDayHeaderFormatter;
    private slicer;
    render(): createElement.JSX.Element;
}

interface CellRenderConfig<RenderProps> {
    generatorName: string;
    customGenerator: CustomContentGenerator<RenderProps>;
    classNameGenerator: ClassNamesGenerator<RenderProps>;
    didMount: DidMountHandler<RenderProps & {
        el: HTMLElement;
    }>;
    willUnmount: WillUnmountHandler<RenderProps & {
        el: HTMLElement;
    }>;
}
interface CellDataConfig<RenderProps> {
    key: string;
    renderProps: RenderProps;
    className?: string;
    attrs?: any;
    innerAttrs?: any;
    colSpan?: number;
    isNavLink?: boolean;
}
interface RowConfig<RenderProps> {
    renderConfig: CellRenderConfig<RenderProps>;
    dataConfigs: CellDataConfig<RenderProps>[];
}
declare function buildDateRowConfigs(...args: Parameters<typeof buildDateRowConfig>): RowConfig<DayHeaderContentArg>[];
declare function buildDateRowConfig(dates: DateMarker[], datesRepDistinctDays: boolean, dateProfile: DateProfile, todayRange: DateRange, dayHeaderFormat: DateFormatter, // TODO: rename to dateHeaderFormat?
context: ViewContext, colSpan?: number): RowConfig<DayHeaderContentArg>;
declare function buildDateRenderConfig(context: ViewContext): CellRenderConfig<DayHeaderContentArg>;
declare function buildDateDataConfigs(dates: DateMarker[], datesRepDistinctDays: boolean, dateProfile: DateProfile, todayRange: DateRange, dayHeaderFormat: DateFormatter, // TODO: rename to dateHeaderFormat?
context: ViewContext, colSpan?: number, keyPrefix?: string): CellDataConfig<DayHeaderContentArg>[];

interface DayGridHeaderRowProps<RenderProps> extends RowConfig<RenderProps> {
    isSticky?: boolean;
    className?: string;
    height?: number;
    colWidth?: number;
    innerHeightRef?: Ref<number>;
    role?: string;
    rowIndex?: number;
}
declare class DayGridHeaderRow<RenderProps extends {
    text: string;
    isDisabled: boolean;
}> extends BaseComponent<DayGridHeaderRowProps<RenderProps>> {
    private innerHeightRefMap;
    private currentInnerHeight?;
    render(): createElement.JSX.Element;
    private handleInnerHeights;
}

declare function buildDayTableModel(dateProfile: DateProfile, dateProfileGenerator: DateProfileGenerator): DayTableModel;
declare function computeColWidth(colCnt: number, colMinWidth: number, viewportWidth: number | undefined): [
    canvasWidth: number | undefined,
    colWidth: number | undefined
];
declare function computeColFromPosition(positionLeft: number, elWidth: number, colWidth: number | undefined, colCnt: number, isRtl: boolean): {
    col: number;
    left: number;
    right: number;
};
declare function getRowEl(rootEl: HTMLElement, row: number): HTMLElement;
declare function getCellEl(rowEl: HTMLElement, col: number): HTMLElement;
declare function createDayHeaderFormatter(explicitFormat: DateFormatter, datesRepDistinctDays: any, dateCnt: any): DateFormatter;

interface DayGridLayoutProps {
    labelId: string | undefined;
    labelStr: string | undefined;
    dateProfile: DateProfile;
    todayRange: DateRange;
    cellRows: DayTableCell[][];
    forPrint: boolean;
    isHitComboAllowed?: (hit0: Hit, hit1: Hit) => boolean;
    className: string;
    headerTiers: RowConfig<{
        text: string;
        isDisabled: boolean;
    }>[];
    fgEventSegs: (DayGridRange & EventRangeProps)[];
    bgEventSegs: (DayGridRange & EventRangeProps)[];
    businessHourSegs: (DayGridRange & EventRangeProps)[];
    dateSelectionSegs: (DayGridRange & EventRangeProps)[];
    eventDrag: EventSegUiInteractionState<DayGridRange> | null;
    eventResize: EventSegUiInteractionState<DayGridRange> | null;
    eventSelection: string;
}
declare class DayGridLayout extends BaseComponent<DayGridLayoutProps> {
    private scrollerRef;
    private rowHeightRefMap;
    private scrollDate;
    render(): createElement.JSX.Element;
    componentDidMount(): void;
    componentDidUpdate(prevProps: DayGridLayoutProps): void;
    componentWillUnmount(): void;
    resetScroll(): void;
    updateScrollY: () => void;
    handleScrollEnd: ({ isUser }: {
        isUser: boolean;
    }) => void;
}

interface DayGridRowsProps {
    dateProfile: DateProfile;
    todayRange: DateRange;
    cellRows: DayTableCell[][];
    forPrint: boolean;
    isHitComboAllowed?: (hit0: Hit, hit1: Hit) => boolean;
    className?: string;
    dayMaxEvents?: number | boolean;
    dayMaxEventRows?: number | boolean;
    fgEventSegs: (DayGridRange & EventRangeProps)[];
    bgEventSegs: (DayGridRange & EventRangeProps)[];
    businessHourSegs: (DayGridRange & EventRangeProps)[];
    dateSelectionSegs: (DayGridRange & EventRangeProps)[];
    eventDrag: EventSegUiInteractionState<DayGridRange> | null;
    eventResize: EventSegUiInteractionState<DayGridRange> | null;
    eventSelection: string;
    colWidth?: number;
    width?: number | string;
    visibleWidth?: number;
    rowHeightRefMap?: RefMap<string, number>;
}
declare class DayGridRows extends DateComponent<DayGridRowsProps> {
    private rootEl;
    private splitBusinessHourSegs;
    private splitBgEventSegs;
    private splitFgEventSegs;
    private splitDateSelectionSegs;
    private splitEventDrag;
    private splitEventResize;
    private rowHeightRefMap;
    private disconnectWidth?;
    render(): createElement.JSX.Element;
    handleRootEl: (rootEl: HTMLDivElement) => void;
    componentDidMount(): void;
    componentWillUnmount(): void;
    queryHit(positionLeft: number, positionTop: number, elWidth: number): Hit;
}
declare function computeRowHeight(visibleWidth: number | undefined, // should INCLUDE any scrollbar width to avoid oscillation
rowCnt: number, isHeightAuto: boolean, forPrint: boolean, options: ViewOptionsRefined): [minHeight: CssDimValue | undefined, isCompact: boolean];

type DayRowRange = SlicedCoordRange;
type DayRowEventRange = DayRowRange & EventRangeProps;
type DayRowEventRangePart = DayRowEventRange & {
    isSlice?: boolean;
    standinFor?: DayRowEventRange;
};

interface DayGridRowProps {
    dateProfile: DateProfile;
    todayRange: DateRange;
    cells: DayTableCell[];
    showDayNumbers: boolean;
    showWeekNumbers?: boolean;
    forPrint: boolean;
    className?: string;
    isCompact?: boolean;
    isTall?: boolean;
    role?: string;
    fgEventSegs: (SlicedCoordRange & EventRangeProps)[];
    bgEventSegs: (SlicedCoordRange & EventRangeProps)[];
    businessHourSegs: (SlicedCoordRange & EventRangeProps)[];
    dateSelectionSegs: (SlicedCoordRange & EventRangeProps)[];
    eventDrag: EventSegUiInteractionState<SlicedCoordRange> | null;
    eventResize: EventSegUiInteractionState<SlicedCoordRange> | null;
    eventSelection: string;
    dayMaxEvents: boolean | number;
    dayMaxEventRows: boolean | number;
    colWidth?: number;
    minHeight?: number | string;
    rootElRef?: Ref<HTMLElement>;
    heightRef?: Ref<number>;
}
declare class DayGridRow extends BaseComponent<DayGridRowProps> {
    private rootEl;
    private headerHeightRefMap;
    private mainHeightRefMap;
    private segHeightRefMap;
    private disconnectHeight?;
    render(): createElement.JSX.Element;
    renderFgSegs(headerHeight: number | undefined, segs: DayRowEventRangePart[], segTops: Map<string, number>, todayRange: DateRange, forcedInvisibleMap: {
        [instanceId: string]: any;
    }, isDragging?: boolean, isResizing?: boolean, isDateSelecting?: boolean): VNode[];
    renderFillSegs(segs: DayRowEventRangePart[], fillType: string): VNode;
    handleRootEl: (rootEl: HTMLElement) => void;
    componentDidMount(): void;
    componentWillUnmount(): void;
    computeFgDims(): [maxMainTop: number | undefined, minMainHeight: number | undefined];
    private handleSegPositioning;
    private getMirrorSegs;
    private getHighlightSegs;
}

export { CellDataConfig, CellRenderConfig, DayGridHeaderRow, DayGridHeaderRowProps, DayGridLayout, DayGridLayoutProps, DayGridRow, DayGridRowProps, DayGridRows, DayGridView, DayTableSlicer, RowConfig, TableDateProfileGenerator, buildDateDataConfigs, buildDateRenderConfig, buildDateRowConfig, buildDateRowConfigs, buildDayTableModel, buildDayTableRenderRange, computeColFromPosition, computeColWidth, computeRowHeight, createDayHeaderFormatter, getCellEl, getRowEl };
